/* qcam - The Quadratic Camera
 *
 * Copyright (C) 2011 - Willem Jan Palenstijn <wjp@usecode.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "config.h"

#include "LAVC.h"
#include <iostream>
#include "SDL.h"


bool LAVC::initialized = false;

void LAVC::initLAVC()
{
	if (initialized) return;

	av_register_all();
	avdevice_register_all();
	initialized = true;
}

bool LAVC::open(const char* filename, SDL_PixelFormat* pformat)
{
	PixelFormat format;
	if (pformat->BitsPerPixel == 32)
		format = PIX_FMT_RGB32;
	else {
		// TODO: Don't assume 565
		format = PIX_FMT_RGB565;
	}

	pFormatCtx = 0;
	pCodecCtx = 0;
	pCodec = 0;
	pFrame = 0;
	pSWSCtx = 0;
	videoStream = -1;
	this->v4l2 = v4l2;
	this->format = format;

	AVInputFormat *inputf = 0;
	inputf = av_find_input_format("video4linux2");
	if (!inputf) {
		std::cerr << "libavformat doesn't support V4L2"
		          << std::endl;
		return false;
	}
	int ret;

#ifdef HAVE_AVFORMAT_OPEN_INPUT
	AVDictionary *options = NULL;
	av_dict_set(&options, "video_size", "1024x768", 0);
	av_dict_set(&options, "framerate", "30", 0);
	ret = avformat_open_input(&pFormatCtx, filename, inputf, &options);
#else
	AVFormatParameters* fp = 0;
	fp = new AVFormatParameters;
	memset(fp, 0, sizeof(fp[0]));
	fp->channels = 0;
	fp->channel = 0;
	fp->width = 1024; // size will/should get adjusted by driver
	fp->height = 768;
	fp->standard = 0;
	fp->prealloced_context = 0;
	fp->time_base = (AVRational){1,30};

	ret = av_open_input_file(&pFormatCtx, filename, inputf, 0, fp);
	delete fp; fp = 0;
#endif
	if (ret != 0) {
		std::cerr << "error opening file: " << ret << std::endl;
		return false; // Couldn't open file
	}

	// TODO: cleanup on errors below

	// Retrieve stream information
#ifdef HAVE_AVFORMAT_FIND_STREAM_INFO
	ret = avformat_find_stream_info(pFormatCtx, 0);
#else
	ret = av_find_stream_info(pFormatCtx);
#endif
	if (ret < 0) {
		std::cerr << "no stream information: " << ret << std::endl;
		return false; // Couldn't find stream information
	}

	// Dump information about file onto standard error
	//dump_format(pFormatCtx, 0, filename.c_str(), false);
#ifdef HAVE_AV_DUMP_FORMAT
	av_dump_format(pFormatCtx, 0, filename, false);
#else
	dump_format(pFormatCtx, 0, filename, false);
#endif

	// Find the first video stream
	for (unsigned int i=0; i < pFormatCtx->nb_streams; i++)
#ifdef HAVE_AVMEDIA_TYPE_VIDEO
		if (pFormatCtx->streams[i]->codec->codec_type == AVMEDIA_TYPE_VIDEO)
#else
		if (pFormatCtx->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO)
#endif
		{
			videoStream=i;
			break;
		}
        if (videoStream == -1) {
		std::cerr << "no video stream found" << std::endl;
		return false; // Didn't find a video stream
	}

	// Get a pointer to the codec context for the video stream
	pCodecCtx = pFormatCtx->streams[videoStream]->codec;

	// Find the decoder for the video stream
	pCodec = avcodec_find_decoder(pCodecCtx->codec_id);
	if (pCodec == NULL) {
		std::cerr << "no codec found" << std::endl;
		return false; // Codec not found
	}

	// Open codec
#ifdef HAVE_AVCODEC_OPEN2
	if (avcodec_open2(pCodecCtx, pCodec, 0) < 0) {
#else 
	if (avcodec_open(pCodecCtx, pCodec) < 0) {
#endif
		std::cerr << "error opening codec" << std::endl;
		return false; // Could not open codec
	}

	msPerFrame = (Uint32)(1000 / av_q2d(pFormatCtx->streams[videoStream]->r_frame_rate));

	// Allocate video frame
	pFrame = avcodec_alloc_frame();

	width = pCodecCtx->width;
	height = pCodecCtx->height;

	std::cout << "Size: " << width << "x" << height << std::endl;

	return true;
}

bool LAVC::getFrame(Uint8* buf, int linesize)
{
	AVPacket packet;
	int frameFinished;
	int ret;

	while ((ret = av_read_frame(pFormatCtx, &packet)) >= 0)
	{
		// Is this a packet from the video stream?
		if (packet.stream_index == videoStream)
		{
			// Decode video frame
#ifdef HAVE_AVCODEC_DECODE_VIDEO2 
			avcodec_decode_video2(pCodecCtx, pFrame,
			                      &frameFinished, &packet);
#else
			avcodec_decode_video(pCodecCtx, pFrame,
			                     &frameFinished, packet.data, packet.size);
#endif
			// Did we get a video frame?
			if (frameFinished)
			{
				if (buf) {
					if (!pSWSCtx) {
						pSWSCtx = sws_getContext(pCodecCtx->width, pCodecCtx->height, pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height, format, SWS_FAST_BILINEAR, 0, 0, 0);
					}
	
					// Convert the image from its native format
					sws_scale(pSWSCtx, pFrame->data,
					          pFrame->linesize, 0,
					          pCodecCtx->height, &buf, &linesize);
				}
				av_free_packet(&packet);
				return true;
			} else {
				// TODO: what do we do here if !frameFinished?
				av_free_packet(&packet);
				return false;
			}
		}
		// Free the packet that was allocated by av_read_frame
		av_free_packet(&packet);
	}
	return false;
}

LAVC::~LAVC()
{
	if (!pFormatCtx) return;

	if (pCodecCtx)
		avcodec_close(pCodecCtx);
	pCodecCtx = 0;
	pCodec = 0;
#ifdef HAVE_AVFORMAT_CLOSE_INPUT
	avformat_close_input(&pFormatCtx);
#else
	av_close_input_file(pFormatCtx);
#endif
	pFormatCtx = 0;
	if (pFrame)
		av_free(pFrame);
	pFrame = 0;
	if (pSWSCtx)
		sws_freeContext(pSWSCtx);
	pSWSCtx = 0;
}

// static
std::vector<std::string> LAVC::findDevices()
{
	std::vector<std::string> devices;

	struct stat s;
	for (int i = 0; i < 64; ++i) {
		char buf[32];
		sprintf(buf, "/dev/video%d", i);
		if (stat(buf, &s) != 0)
			continue;
		bool ok = true;

#if 0
		printf("Trying %s\n", buf);

		SDL_Surface* screen = SDL_GetVideoSurface();
		PixelFormat format;
		if (screen->format->BitsPerPixel == 32)
			format = PIX_FMT_RGB32;
		else
			format = PIX_FMT_RGB565;

		LAVC* l = new LAVC();
		ok = l->open(buf, true, format);
		delete l;
		if (ok) {
			// Some video drivers seem to become thoroughly confused if
			// we re-open them immediately after closing them.
			sleep(2);
		}
#endif

		if (ok) {
			devices.push_back(buf);
			printf("Device %s seems ok\n", buf);
		}
	}

	return devices;
}

