/* qcam - The Quadratic Camera
 *
 * Copyright (C) 2011 - Willem Jan Palenstijn <wjp@usecode.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "VI.h"
#include "lib_win32/videoInput.h"

videoInput* VI::vi;

VI::VI()
{
	initVI();

	framebuf = 0;
	device = -1;
}

VI::~VI()
{
	if (device >= 0) {
		vi->stopDevice(device);
		device = -1;
	}

	delete[] framebuf;
	framebuf = 0;
}

//static
void VI::initVI()
{
	if (vi) return;

	vi = new videoInput();
}

//static
void VI::quit()
{
	delete vi;
	vi = 0;
}

bool VI::open(const char* filename, SDL_PixelFormat* format)
{
	// HACK: Due to linux needing a device name, we store cam numbers as strings
	int c = atoi(filename);

	if (device >= 0)
		vi->stopDevice(device);

	bool ok = vi->setupDevice(c, 1024, 768);

	if (!ok)
		return false;

	device = c;

	width = vi->getWidth(device);
	height = vi->getHeight(device);

	pformat = *format;

	delete[] framebuf;
	framebuf = new Uint8[vi->getSize(device)];

	return true;
}

static Uint8 broken_frame[] = { 0x01, 0x00, 0x01, 0xFC, 0xFA, 0xFC, 0x49, 0x47, 0x49, 0x94, 0x92, 0x94, 0x7A, 0x78, 0x7A };

bool VI::getFrame(Uint8* buf, int linesize)
{
	if (device < 0)
		return false;

	if (!vi->isFrameNew(device))
		return true;

	bool ok = vi->getPixels(device, framebuf, false, true);
	if (!ok)
		return false;

	// HACK: My webcam seems to return a gray grid every once in a while.
	// Skip those frames.
	if (memcmp(framebuf, broken_frame, sizeof(broken_frame)) == 0)
		return true;

	// videoInput stores its output as packed 24 bit RGB, so we have
	// to do yet another conversion here.


	if (pformat.BytesPerPixel == 2) { 

		for(int y = 0; y < height; ++y)
		{
			Uint16* pixels = (Uint16*)(buf + y*linesize);

			for(int x = 0; x < width; ++x)
			{
				Uint32* rgbptr = (Uint32*)(framebuf + y*width*3+3*x);
				Uint32 c = *rgbptr;
				Uint8 b = c;
				Uint8 g = (c >> 8);
				Uint8 r = (c >> 16);
				*pixels++ = SDL_MapRGB(&pformat, r, g, b);
			}
		}

	} else {

		for(int y = 0; y < height; ++y)
		{
			Uint32* pixels = (Uint32*)(buf + y*linesize);

			for(int x = 0; x < width; ++x)
			{
				Uint32* rgbptr = (Uint32*)(framebuf + y*width*3+3*x);
				Uint32 c = *rgbptr;
				Uint8 b = c;
				Uint8 g = c >> 8;
				Uint8 r = c >> 16;
				*pixels++ = SDL_MapRGB(&pformat, r, g, b);
			}
		}

	}

}

// static
std::vector<std::string> VI::findDevices()
{
	initVI();

	std::vector<std::string> devices;

	int numDevices = vi->listDevices();

	for (int i = 0; i < numDevices; ++i) {
		char buf[32];
		sprintf(buf, "%d", i);
		devices.push_back(buf);
	}

	return devices;
}
