/* qcam - The Quadratic Camera
 *
 * Copyright (C) 2011 - Bart de Smit <desmit@math.leidenuniv.nl>
 *                  and Willem Jan Palenstijn <wjp@usecode.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "SDL.h"

#ifndef WIN32
#include "LAVC.h"
#else
#include "VI.h"
#endif

#ifdef WIN32
typedef unsigned int uint;
#endif

#include <iostream>
#include <vector>
#include <string>
#include <sys/stat.h>
#include <cmath>

#define TEXTURE_XEXP 14
#define TEXTURE_YEXP 14

#define TEXTURE_WIDTH (1<<TEXTURE_XEXP)
#define TEXTURE_HEIGHT (1<<TEXTURE_YEXP)
#define TEXTURE_XMASK (TEXTURE_WIDTH-1)
#define TEXTURE_YMASK (TEXTURE_HEIGHT-1)

static bool fullscreen = true;

unsigned int cam_num=0;

int dump_count=0;
char dump_name [20]; 

int screenw;
int screenh;

// For storing the last used window size and fullscreen size
int window_w, window_h;
int fullscreen_w, fullscreen_h;

int viewer_size=0;
int zoom_type=0;
Uint32* transform=0;
int video_width, video_height;
VideoPlayer* lavc = 0;
Uint8* video_buf;

bool paused=false;

std::vector<std::string> devices;

bool loadTransform()
{
	SDL_Surface* screen = SDL_GetVideoSurface();
	int w = screen->w;
	int h = screen->h;

	delete[] transform;
	transform = new Uint32[w*h];
	if (!transform) {
		std::cerr << "Out of memory while building transform." << std::endl;
		return false;
	}

	int x,y;
	double video_aspect= ((double)(video_width))/video_height;
	double screen_aspect= ((double)(w))/h;
	int viewer_w=224*viewer_size;
	int hairs_l=viewer_w/12;
	int hairs_s=6;
	int viewer_h=(int)(viewer_w/video_aspect);
	double co=cos(atan2(video_height,video_width)/2);
	double zoom= sqrt(1+co*co)/sqrt(sqrt(1+ video_aspect*video_aspect));
	double sam=1; if (screen_aspect <1 ) {sam=screen_aspect;}
	if (zoom_type==0){zoom=zoom*sam;}
	if (zoom_type==1){zoom=zoom*screen_aspect/sam;}
	if (zoom_type==2){zoom=1/sam*screen_aspect/sqrt(video_aspect);}
	for(int iy=0;iy<h;iy++)
	for(int ix=0;ix<w;ix++) {
		if ((ix>=w-viewer_w)&&(iy<viewer_h)) {
			if (((ix+1)/2 == w/2-viewer_w/4 && -hairs_l < viewer_h/2-iy && hairs_l > viewer_h/2-iy &&
			   not(-hairs_s < viewer_h/2-iy && hairs_s >= viewer_h/2-iy)) ||
			   ((iy+1)/2==viewer_h/4 && -hairs_l < viewer_w/2-w+ix && hairs_l > viewer_w/2-w+ix &&
			   not(-hairs_s <= viewer_w/2-w+ix && hairs_s > viewer_w/2-w+ix))) {
				x = video_width; y = 0;
			} else {
				x=(uint)((viewer_w-w+ix)*video_width/viewer_w);
				y=(uint)(iy*video_height/viewer_h);
			}
		} else {
			double fx=(double(ix-w/2)/double(w/2))*zoom;
			double fy=(double(h/2-iy)/double(w/2))*zoom;
			if ((zoom_type==2) && (fx*fx+fy*fy>=1/video_aspect)) {
				x = video_width+2; y = 0;
			} else {
				double gx=fx*fx-fy*fy;
				double gy=fy*fx*2;
				x=(uint)((video_width/2+video_width*(gx)/2));
				y=(uint)((video_height/2-video_width*(gy)/2));
				if (x < 0 || x >= video_width || y < 0
					|| y >= video_height) {
						x = video_width+2; y = 0; 
				}
			}
		}
		transform[iy*w+ix]=(Uint32)(x) + (Uint32)(y*TEXTURE_WIDTH);
		//fprintf(stderr, "coords %d %d   --- %d %d   \n", ix,iy, x, y);
	}
	return true;
}

bool render()
{
	if (!lavc) {
		// No camera open at the moment
		return false;
	}

	SDL_Surface* screen = SDL_GetVideoSurface();

	int dh = screen->h, dw = screen->w;
	int dy = 0, dx = 0;

	int ey=dy+dh;
	if (!video_buf) {
		video_buf = new Uint8[screen->format->BytesPerPixel*
		                  video_width*video_height];
		if (!video_buf) {
			std::cerr << "Out of memory while allocating video_buf"
			          << std::endl;
			return false;
		}
	}

	Uint8* pix = video_buf;

	int linesize = screen->format->BytesPerPixel*video_width;
	lavc->getFrame(pix, linesize);

	if (paused) {
		return true;
	}

	Uint32 red = SDL_MapRGB(screen->format, 255, 0, 0);

	SDL_LockSurface(screen);

	if (screen->format->BytesPerPixel == 2) { 
		for(int y=dy;y<ey;y++)
		{
			Uint16* pixels = (Uint16*)((Uint8*)screen->pixels +
			            y*screen->pitch) + dx;

			for(int x=dx;x<dx+dw; x++)
			{
				Uint32 videocoords;
				videocoords=transform[x+y*screenw];
				uint ix = videocoords&TEXTURE_XMASK;
				uint iy = (videocoords/TEXTURE_WIDTH)&TEXTURE_YMASK;
				if (ix<(uint)video_width) {
					Uint16* rgbptr = (Uint16*)video_buf;
					rgbptr += iy*video_width+ix;
					*pixels = *rgbptr;
				} else if (ix==(uint)video_width)
					*pixels = red;
				else 
					*pixels = 0;

				pixels++;
			}
		}

	} else {
		for(int y=dy;y<ey;y++)
		{
			Uint32* pixels = (Uint32*)((Uint8*)screen->pixels +
			                                 y*screen->pitch) + dx;

			for(int x=dx;x<dx+dw; x++)
			{
				Uint32 videocoords;
				videocoords=transform[x+y*screenw];
				uint ix = videocoords&TEXTURE_XMASK;
				uint iy = (videocoords/TEXTURE_WIDTH)&TEXTURE_YMASK;
				if (ix<(uint)video_width) {
					Uint32* rgbptr = (Uint32*)video_buf;
					rgbptr += iy*video_width+ix;
					*pixels = *rgbptr;
				} else if (ix==(uint)video_width)
					*pixels = red;
				else 
					*pixels = 0;

				pixels++;
			}
		}

	}

	SDL_UnlockSurface(screen);
	SDL_UpdateRect(screen, 0, 0, 0, 0);
	return true;
}

void liveDelay(unsigned int msecs)
{
	bool old_paused = paused;
	paused = true;
	Uint32 start = SDL_GetTicks();
	while (SDL_GetTicks() < start + msecs) {
		render();
		SDL_Delay(10);
	}
	paused = old_paused;
}

bool start()
{
	SDL_Surface* screen = SDL_GetVideoSurface();

	SDL_FillRect(screen, 0, SDL_MapRGB(screen->format, 128, 128, 128));
	SDL_UpdateRect(screen, 0, 0, 0, 0);

	if (lavc) {
		delete lavc;
		// Some video drivers seem to become thoroughly confused if
		// we re-open them immediately after closing them.
		SDL_Delay(2000);
	}

	delete[] video_buf;
	video_buf = 0;

#ifndef WIN32
	lavc = new LAVC();
#else
	lavc = new VI();
#endif
	if (!lavc->open(devices[cam_num].c_str(), screen->format)) {
		delete lavc;
		lavc = 0;
		SDL_FillRect(screen, 0, 0);
		SDL_UpdateRect(screen, 0, 0, 0, 0);

		return false;
	}

	lavc->getSize(video_width, video_height);

	if (!loadTransform()) {
		SDL_FillRect(screen, 0, 0);
		SDL_UpdateRect(screen, 0, 0, 0, 0);

		std::cerr << "Error computing transform" << std::endl;
		return false;
	}

	SDL_FillRect(screen, 0, 0);

	render();

	return true;
}

bool setVideoMode()
{
	SDL_Surface* screen;
	if (fullscreen) {
		screen = SDL_SetVideoMode(fullscreen_w,fullscreen_h,
		                  0, SDL_SWSURFACE | SDL_FULLSCREEN);
		SDL_ShowCursor(SDL_DISABLE);

	} else {
		screen = SDL_SetVideoMode(window_w, window_h,
		                  0, SDL_SWSURFACE | SDL_RESIZABLE);
		SDL_ShowCursor(SDL_ENABLE);
	}

	if (!screen)
		return false;

	SDL_WM_SetCaption("Quadratic Camera", "");


	screenw = screen->w;
	screenh = screen->h;

	if (fullscreen) {
		fullscreen_w = screenw;
		fullscreen_h = screenh;
	} else {
		window_w = screenw;
		window_h = screenh;
	}

	return true;
}

bool toggleFullscreen()
{
	fullscreen = !fullscreen;
	bool ok = setVideoMode();
	if (!ok)
		return false;

	// Resolution and/or colordepth may have changed
	return loadTransform();
}

void cmdlineVersion()
{
	std::cout << "qcam version 1.0, Copyright (c) 2011 Bart de Smit, Willem Jan Palenstijn" << std::endl;
}

void cmdlineHelp()
{
	std::cout << "qcam [--fs|--nofs] [--device <dev>] [--size <width> <height>] [--version]" << std::endl;
}

int main(int argc, char* argv[])
{
	bool single_device = false;

	int argw = -1;
	int argh = -1;

	for (int i = 1; i < argc; ++i) {
		if (strcmp(argv[i], "--version") == 0 ||
		             strcmp(argv[i], "-v") == 0) {
			cmdlineVersion();
			return 0;
		} else if (strcmp(argv[i], "--help") == 0 ||
		             strcmp(argv[i], "-h") == 0) {
			cmdlineHelp();
			return 0;
		} else if (strcmp(argv[i], "--fs") == 0)
			fullscreen = true;
		else if (strcmp(argv[i], "--nofs") == 0)
			fullscreen = false;
		else if (strcmp(argv[i], "--device") == 0) {
			if (i+1 == argc) {
				cmdlineHelp();
				return 1;
			}
			devices.push_back(argv[++i]);
			single_device = true;
		} else if (strcmp(argv[i], "--size") == 0) {
			if (argc < i+3) {
				cmdlineHelp();
				return 1;
			}
			argw = atoi(argv[++i]);
			argh = atoi(argv[++i]);
		}
	}

	SDL_Init(SDL_INIT_VIDEO);

	fullscreen_w = -1;
	fullscreen_h = -1;
	window_w = -1;
	window_h = -1;

	if (argw >= 0 && argh >= 0) {
		if (fullscreen) {
			fullscreen_w = argw;
			fullscreen_h = argh;
		} else {
			window_w = argw;
			window_h = argh;
		}
	}

	if (fullscreen_w == -1) {
		if (fullscreen) {
			fullscreen_w = 0;
			fullscreen_h = 0;
		} else {
			const SDL_VideoInfo* info = SDL_GetVideoInfo();
			fullscreen_w = info->current_w;
			fullscreen_h = info->current_h;
		}
	}

	if (window_w == -1) {
		window_w = 800;
		window_h = 600;
	}

	bool ok = setVideoMode();


	if (!ok) {
		std::cerr << "Failed to set video mode: " << SDL_GetError() << std::endl;
		SDL_Quit();
		return 1;
	}

	if (!single_device) {
#ifndef WIN32
		devices = LAVC::findDevices();
#else
		devices = VI::findDevices();
#endif
	}

	if (devices.empty()) {
		std::cerr << "No camera device /dev/video* found" << std::endl;
		SDL_Quit();
		return 1;
	}

	cam_num = 0;

	ok = start();
	if (!ok && single_device) {
		SDL_Quit();
		return 1;
	}

	bool quit = false;

	while (!quit) {
		SDL_Event event;
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
				case SDLK_s:
				{
					struct stat info;
					do {
						dump_count++;
						sprintf(dump_name,"screen_%03d.bmp",dump_count);
					} while (!stat(dump_name,&info));

					SDL_Surface* screen = SDL_GetVideoSurface();
					SDL_Surface* backup = SDL_CreateRGBSurface(SDL_SWSURFACE, screen->w, screen->h, screen->format->BitsPerPixel, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);
					SDL_BlitSurface(screen, 0, backup, 0);
					SDL_SaveBMP(screen,dump_name);

					// Flash screen for about 0.1s as a 'shutter effect'
					SDL_FillRect(screen, 0, 0);
					SDL_UpdateRect(screen, 0, 0, 0, 0);
					liveDelay(50);
					// Re-show the captured image as both a preview and
					// for the camera to skip the black image.
					SDL_BlitSurface(backup, 0, screen, 0);
					SDL_UpdateRect(screen, 0, 0, 0, 0);
					SDL_FreeSurface(backup);
					liveDelay(400);
					break;
				}
				case SDLK_c:
					if (devices.size() > 1) {
						cam_num++;
						if (cam_num >= devices.size())
							cam_num = 0;
						start();
					}
					break;
				case SDLK_f:
				{
					bool ok = toggleFullscreen();
					if (!ok) {
						std::cerr << "Failed to set video mode: " << SDL_GetError() << std::endl;
						SDL_Quit();
						return 1;
					}
					break;
				}
				case SDLK_RETURN:
					if (event.key.keysym.mod & (KMOD_LALT | KMOD_RALT)) {
						bool ok = toggleFullscreen();
						if (!ok) {
							std::cerr << "Failed to set video mode: " << SDL_GetError() << std::endl;
							SDL_Quit();
							return 1;
						}
					}
					break;
				case SDLK_p:
					paused=!paused;
					break;
				case SDLK_z:
					zoom_type=(zoom_type+1)%3;
					loadTransform();
					break;
				case SDLK_v:
					viewer_size=(viewer_size+1)&3;
					loadTransform();
					break;
				case SDLK_ESCAPE:
					quit = true;
				default:
					break;
				}
				break;
			case SDL_QUIT:
				quit = true;
				break;
			case SDL_VIDEORESIZE:
			{
				window_w = event.resize.w;
				window_h = event.resize.h;
				// TODO: Do we want to fix aspect ratio?
				bool ok = setVideoMode();
				if (!ok) {
					std::cerr << "Failed to set video mode: " << SDL_GetError() << std::endl;
					SDL_Quit();
					return 1;
				}

				loadTransform();
				break;
			}
			default:
				break;
			}
		}
		render();
		SDL_Delay(10);
	}
	SDL_Quit();
	return 0;
}
